const chai = require('chai');
const chaiHttp = require('chai-http');
const fs = require('fs');
const path = require('path');
const app = require('../index');
const expect = chai.expect;
const mongoose = require('mongoose');
const { GraphQLClient } = require('graphql-request');

chai.use(chaiHttp);
const endpoint = 'http://localhost:8080/graphql';
const client = new GraphQLClient(endpoint);
let agent1AuthToken;
let agent2AuthToken;
let nonAgentAuthToken;
let adId;

describe('Ad Routes', () => {
    before(async () => {
        try {
            // Login as 'hajar'
            const hajarLoginResponse = await chai
                .request(app)
                .post('/user/login')
                .send({ username: 'hajar', password: 'hajar' });

            agent1AuthToken = hajarLoginResponse.body.token;

            // Login as 'ilham'
            const ilhamLoginResponse = await chai
                .request(app)
                .post('/user/login')
                .send({ username: 'ilham', password: 'ilham' });

            agent2AuthToken = ilhamLoginResponse.body.token;

            const nonAgentLoginResponse = await chai
                .request(app)
                .post('/user/login')
                .send({ username: 'manal', password: 'manal' });

            nonAgentAuthToken = nonAgentLoginResponse.body.token;

            // Create an ad to get the adId for some tests (using 'hajar' user)
            const adResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Test Ad',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });

            adId = adResponse.body._id;
        } catch (error) {
            console.error('Error obtaining authentication tokens:', error);
        }
    });

    describe('POST /ads', () => {
        it('should add a new ad', async () => {
            const response = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Test Ad',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });
            expect(response.body).to.have.property('_id').to.be.a('string');
        });

    const backupFolderPath = path.join(__dirname, '..', 'backup');

        it('should add a new ad with photos', async () => {
            const response = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .attach('photos', path.join(backupFolderPath, 'img1.jpg'))
                .attach('photos', path.join(backupFolderPath, 'img2.jpg'))
                .field('title', 'Test Ad 2')
                .field('propertyType', 'À la vente')
                .field('publicationStatus', 'Publiée')
                .field('propertyStatus', 'Disponible')
                .field('description', 'teest.')
                .field('price', 80000)
                .field('availabilityDate', '2024-11-01');

            expect(response).to.have.status(200);
            expect(response.body).to.have.property('_id').to.be.a('string');
            adId = response.body._id;
        });

        it('should not allow a non-authorized user (non-agent) to add a new ad', async () => {
            // Attempt to add a new ad as a non-agent user
            const response = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${nonAgentAuthToken}`)
                .send({
                    title: 'Test Ad',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });

            expect(response).to.have.status(403);
            expect(response.body).to.have.property('message').to.equal('User non autorisé');

            const logoutResponse = await chai
                .request(app)
                .post('/user/logout')
                .set('Authorization', `Bearer ${nonAgentAuthToken}`);

            expect(logoutResponse).to.have.status(200);
        });

    });

    describe('PUT /ads/:id', () => {
        it('should update an existing ad with the agent user (hajar)', async () => {
            // Create a new ad to update
            const createResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Test Ad to update',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });

            const adIdToUpdate = createResponse.body._id;

            // Update the ad with the agent user (hajar)
            const updateResponse = await chai
                .request(app)
                .put(`/ad/${adIdToUpdate}`)
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Updated Test Ad',
                    description: 'Updated description',
                    price: 90000,
                });
            expect(updateResponse).to.have.status(200);
            expect(updateResponse.body).to.have.property('_id').to.equal(adIdToUpdate);
            expect(updateResponse.body).to.have.property('title').to.equal('Updated Test Ad');
            expect(updateResponse.body).to.have.property('description').to.equal('Updated description');
            expect(updateResponse.body).to.have.property('price').to.equal(90000);

        });
    });
    describe
    ('POST /ad/:id/photos', () => {
        it('should add photos to an existing ad', async () => {
            const imagePath = path.join(__dirname, '..', 'backup', 'img3.jpg');

            const response = await chai
                .request(app)
                .post(`/ad/${adId}/photos`)
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .attach('photos', fs.readFileSync(imagePath), 'img3.jpg');

            expect(response).to.have.status(200);
            expect(response.body).to.have.property('_id');
            expect(response.body.photos).to.be.an('array');
        });
    });

    describe('GET /ads', () => {
        it('should get all ads', async () => {
            const response = await chai
                .request(app)
                .get('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`);

            expect(response).to.have.status(200);
            expect(response.body).to.be.an('array');
        });

    });
    describe('GET /ad/:id', () => {
        it('should get an ad by ID', async () => {
            const response = await chai
                .request(app)
                .get(`/ad/${adId}`)
                .set('Authorization', `Bearer ${agent1AuthToken}`);

            expect(response).to.have.status(200);
            expect(response.body).to.be.an('object');
            expect(response.body).to.have.property('_id').to.equal(adId);
        });

        it('should return 404 for non-existent ad ID', async () => {
            const nonExistentAdId = new mongoose.Types.ObjectId();
            const response = await chai
                .request(app)
                .get(`/ad/${nonExistentAdId}`)
                .set('Authorization', `Bearer ${agent1AuthToken}`);
            expect(response).to.have.status(404);
            expect(response.body).to.have.property('message').to.equal('Cette annonce est introuvable ou non disponible.');
        });
    });

    describe('DELETE /ad/:id', () => {
        it('should delete an existing ad', async () => {
            const response = await chai
                .request(app)
                .delete(`/ad/${adId}`)
                .set('Authorization', `Bearer ${agent1AuthToken}`);

            expect(response).to.have.status(200);
            expect(response.body).to.have.property('message').to.equal('Annonce supprimée avec succès');
        });

        it('should return 404 for deleting non-existent ad ID', async () => {
            const nonExistentAdId = new mongoose.Types.ObjectId();
            const response = await chai
                .request(app)
                .delete(`/ad/${nonExistentAdId}`)
                .set('Authorization', `Bearer ${agent1AuthToken}`);

            expect(response).to.have.status(404);
            expect(response.body).to.have.property('message').to.equal('Ad not found.');
        });
    });

    describe('PUT /ad/:id/ask', () => {
        it('should allow a user to ask a question about an ad', async () => {
            const adResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Test Ad for Asking Questions',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description for asking questions.',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });

            const adIdForAskingQuestion = adResponse.body._id;

            const response = await chai
                .request(app)
                .put(`/ad/${adIdForAskingQuestion}/ask`)
                .set('Authorization', `Bearer ${nonAgentAuthToken}`)
                .send({
                    question: 'This is a test question.',
                });

            expect(response).to.have.status(200);
            expect(response.body).to.have.property('questions').to.be.an('array');
            expect(response.body.questions[0]).to.have.property('user').to.equal('manal');
            expect(response.body.questions[0]).to.have.property('question').to.equal('This is a test question.');
        });
    });

    describe('PUT /ad/:id/question/:questionId/answer', () => {
        it('should allow the owner to answer a question about their ad', async () => {
            const adResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Test Ad for Answering Questions',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description for answering questions.',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });

            const adIdForAnsweringQuestion = adResponse.body._id;

            const askQuestionResponse = await chai
                .request(app)
                .put(`/ad/${adIdForAnsweringQuestion}/ask`)
                .set('Authorization', `Bearer ${nonAgentAuthToken}`)
                .send({
                    question: 'This is a test question.',
                });

            const questionId = askQuestionResponse.body.questions[0]._id;

            const answerResponse = await chai
                .request(app)
                .put(`/ad/${adIdForAnsweringQuestion}/question/${questionId}/answer`)
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    answer: 'This is a test answer.',
                });
            expect(answerResponse).to.have.status(200);
            expect(answerResponse.body).to.have.property('questions').to.be.an('array');
            expect(answerResponse.body.questions[0].answers[0]).to.have.property('answer').to.equal('This is a test answer.');
        });


        it('should not allow a non-owner to answer a question about an ad', async () => {
            const adResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Test Ad for Non-Owner Answering',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description for non-owner answering.',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });

            const adIdForNonOwnerAnswering = adResponse.body._id;

            const askQuestionResponse = await chai
                .request(app)
                .put(`/ad/${adIdForNonOwnerAnswering}/ask`)
                .set('Authorization', `Bearer ${nonAgentAuthToken}`)
                .send({
                    question: 'This is a test question.',
                });

            const questionId = askQuestionResponse.body.questions[0]._id;

            const answerResponse = await chai
                .request(app)
                .put(`/ad/${adIdForNonOwnerAnswering}/question/${questionId}/answer`)
                .set('Authorization', `Bearer ${agent2AuthToken}`)
                .send({
                    answer: 'This is a test answer.',
                });
            expect(answerResponse).to.have.status(403);
            expect(answerResponse.body).to.have.property('message').to.equal('User non autorisé.');
        });
    });

    describe('Query graphQl : getAds and getAd By Id', () => {
        it('should return a list of ads', async () => {
            const query = `
                query {
                    getAds(token: "${agent1AuthToken}") {
                        _id
                        title
                        propertyType
                        publicationStatus
                        propertyStatus
                        description
                        price
                        availabilityDate
                        photos
                        userName
                        questions {
                            user
                            question
                        }
                    }
                }
            `;

            const response = await chai
                .request(app)
                .post('/graphql')
                .send({ query });

            expect(response).to.have.status(200);
            expect(response.body).to.have.property('data');
            expect(response.body.data).to.have.property('getAds');
        });
        it('should return details of a specific ad', async () => {
            const query = `
                query {
                    getAdById(adId: "${adId}", token: "${agent2AuthToken}") {
                        _id
                        title
                        propertyType
                        publicationStatus
                        propertyStatus
                        description
                        price
                        availabilityDate
                        photos
                        userName
                        questions {
                            user
                            question
                        }
                    }
                }
            `;

            const response = await chai
                .request(app)
                .post('/graphql')
                .send({ query });

            expect(response).to.have.status(200);
            expect(response.body).to.have.property('data');
            expect(response.body.data).to.have.property('getAdById');
        });
    });

    describe('Mutation: addAd', () => {
        it('Adds a new ad and returns the correct fields', async () => {
            const variables = {
                input: {
                    title: 'Test Ad 2',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'teest.',
                    price: 80000,
                    availabilityDate: '2024-11-01',
                },
            };
            const headers = {
                Authorization: `Bearer ${agent1AuthToken}`,
            };
            const mutation = `
              mutation AddAd($input: AdInput!) {
                addAd(input: $input) {
                  title
                  propertyType
                  publicationStatus
                  propertyStatus
                  description
                  price
                  availabilityDate
                }
              }
            `;

            const response = await client.request(mutation, variables, headers);
            expect(response.addAd).to.deep.equal({
                title: 'Test Ad 2',
                propertyType: 'À la vente',
                publicationStatus: 'Publiée',
                propertyStatus: 'Disponible',
                description: 'teest.',
                price: 80000,
                availabilityDate: '1730419200000',
            });
        });
    });
    describe('Mutation: updateAd', () => {
        it('Updates an existing ad and returns the correct fields', async () => {
            const adToUpdate = {
                title: 'Ad to Update',
                propertyType: 'À la vente',
                publicationStatus: 'Publiée',
                propertyStatus: 'Disponible',
                description: 'Test description',
                price: 80000,
                availabilityDate: '2024-12-01',
            };

            const createResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send(adToUpdate);

            const adIdToUpdate = createResponse.body._id;
            const updatedAd = {
                title: 'Updated Ad',
                description: 'Updated description',
                price: 90000,
            };

            const updateResponse = await chai
                .request(app)
                .put(`/ad/${adIdToUpdate}`)
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send(updatedAd);

            expect(updateResponse).to.have.status(200);
            expect(updateResponse.body).to.have.property('_id').to.equal(adIdToUpdate);
            expect(updateResponse.body).to.have.property('title').to.equal(updatedAd.title);
            expect(updateResponse.body).to.have.property('description').to.equal(updatedAd.description);
            expect(updateResponse.body).to.have.property('price').to.equal(updatedAd.price);
        });

        it('Does not allow a non-owner to update an ad', async () => {
            const adToUpdate = {
                title: 'Ad to Update',
                propertyType: 'À la vente',
                publicationStatus: 'Publiée',
                propertyStatus: 'Disponible',
                description: 'Test description',
                price: 80000,
                availabilityDate: '2024-12-01',
            };
            const createResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send(adToUpdate);

            const adIdToUpdate = createResponse.body._id;

            const updateResponse = await chai
                .request(app)
                .put(`/ad/${adIdToUpdate}`)
                .set('Authorization', `Bearer ${agent2AuthToken}`)
                .send({
                    title: 'Attempted Update',
                    description: 'Attempted update description',
                    price: 95000,
                });

            expect(updateResponse).to.have.status(403);
            expect(updateResponse.body).to.have.property('message').to.equal('User non autorisé');
        });
    });
    describe('Mutation: deleteAd', () => {
        it('Deletes an existing ad', async () => {
            const createResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Ad to Delete',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });

            const adIdToDelete = createResponse.body._id;
            const deleteResponse = await chai
                .request(app)
                .delete(`/ad/${adIdToDelete}`)
                .set('Authorization', `Bearer ${agent1AuthToken}`);

            expect(deleteResponse).to.have.status(200);
            expect(deleteResponse.body).to.have.property('message').to.equal('Annonce supprimée avec succès');
        });

        it('Does not allow a non-owner to delete an ad', async () => {
            // Create a new ad to delete
            const createResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Ad to Delete',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });

            const adIdToDelete = createResponse.body._id;

            const deleteResponse = await chai
                .request(app)
                .delete(`/ad/${adIdToDelete}`)
                .set('Authorization', `Bearer ${agent2AuthToken}`);
            expect(deleteResponse).to.have.status(403);
            expect(deleteResponse.body).to.have.property('message').to.equal('User non autorisé');
        });
    });

    describe('Mutation: askQuestion and answerQuestion', () => {
        it('Allows a user to ask a question about an ad', async () => {
            const createResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Ad for Asking Questions',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description for asking questions.',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });

            const adIdForAskingQuestion = createResponse.body._id;

            const askQuestionResponse = await chai
                .request(app)
                .put(`/ad/${adIdForAskingQuestion}/ask`)
                .set('Authorization', `Bearer ${nonAgentAuthToken}`)
                .send({
                    question: 'This is a test question.',
                });

            expect(askQuestionResponse).to.have.status(200);
            expect(askQuestionResponse.body).to.have.property('questions').to.be.an('array');
            expect(askQuestionResponse.body.questions[0]).to.have.property('user').to.equal('manal');
            expect(askQuestionResponse.body.questions[0]).to.have.property('question').to.equal('This is a test question.');
        });

        it('Allows the owner to answer a question about their ad', async () => {
            const createResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Ad for Answering Questions',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description for answering questions.',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });

            const adIdForAnsweringQuestion = createResponse.body._id;

            const askQuestionResponse = await chai
                .request(app)
                .put(`/ad/${adIdForAnsweringQuestion}/ask`)
                .set('Authorization', `Bearer ${nonAgentAuthToken}`)
                .send({
                    question: 'This is a test question.',
                });

            const questionId = askQuestionResponse.body.questions[0]._id;

            const answerResponse = await chai
                .request(app)
                .put(`/ad/${adIdForAnsweringQuestion}/question/${questionId}/answer`)
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    answer: 'This is a test answer.',
                });

            expect(answerResponse).to.have.status(200);
            expect(answerResponse.body).to.have.property('questions').to.be.an('array');
            expect(answerResponse.body.questions[0].answers[0]).to.have.property('answer').to.equal('This is a test answer.');
        });

        it('Does not allow a non-owner to answer a question about an ad', async () => {
            const createResponse = await chai
                .request(app)
                .post('/ad')
                .set('Authorization', `Bearer ${agent1AuthToken}`)
                .send({
                    title: 'Ad for Answering Questions',
                    propertyType: 'À la vente',
                    publicationStatus: 'Publiée',
                    propertyStatus: 'Disponible',
                    description: 'Test description for answering questions.',
                    price: 80000,
                    availabilityDate: '2024-12-01',
                });

            const adIdForAnsweringQuestion = createResponse.body._id;

            const askQuestionResponse = await chai
                .request(app)
                .put(`/ad/${adIdForAnsweringQuestion}/ask`)
                .set('Authorization', `Bearer ${nonAgentAuthToken}`)
                .send({
                    question: 'This is a test question.',
                });

            const questionId = askQuestionResponse.body.questions[0]._id;

            const answerResponse = await chai
                .request(app)
                .put(`/ad/${adIdForAnsweringQuestion}/question/${questionId}/answer`)
                .set('Authorization', `Bearer ${agent2AuthToken}`)
                .send({
                    answer: 'This is a test answer.',
                });

            expect(answerResponse).to.have.status(403);
            expect(answerResponse.body).to.have.property('message').to.equal('User non autorisé.');
        });
    });

    // Cleanup after tests
    after(async () => {
        try {
            // Delete the ad if adId is defined
            if (adId) {
                await chai
                    .request(app)
                    .delete(`/ad/${adId}`)
                    .set('Authorization', `Bearer ${agent1AuthToken}`);
            }

            const photo1Path = path.join(__dirname, '..', 'uploads', 'img1.jpg');
            const photo2Path = path.join(__dirname, '..', 'uploads', 'img2.jpg');
            const photo3Path = path.join(__dirname, '..', 'uploads', 'img3.jpg');
            // Delete the photos
            if (fs.existsSync(photo1Path)) {
                fs.unlinkSync(photo1Path);
            }

            if (fs.existsSync(photo2Path)) {
                fs.unlinkSync(photo2Path);
            }

            if (fs.existsSync(photo3Path)) {
                fs.unlinkSync(photo3Path);
            }
        } catch (error) {
            console.error('Error cleaning up:', error);
        }
    });
});
